@@ -76,6 +76,8 @@ static void help() {
 }
 
 static void xwrite(const void * data, size_t size, size_t len, FILE * des) {
+    if (len == 0 || size == 0)
+        return;
     if (fwrite(data, size, len, des) != len) {
         fprintf(stderr, "Write error: %s\n", strerror(errno));
         exit(1);
@@ -141,7 +143,8 @@ static void close_out_file(FILE * des) {
     }
 }
 
-static int process(FILE * input_des, FILE * output_des, int mode, int block_size, int workers, int verbose, char * file_name) {
+static int process(FILE * input_des, FILE * output_des, int mode, int block_size, int workers, int verbose,
+                   char * file_name) {
     uint64_t bytes_read = 0, bytes_written = 0;
 
     if ((mode == MODE_ENCODE && isatty(fileno(output_des))) ||
@@ -150,6 +153,9 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
         return 1;
     }
 
+    // Reset errno after the isatty() call.
+    errno = 0;
+
     u8 byteswap_buf[4];
 
     switch (mode) {
@@ -214,6 +220,8 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
                 read_count = xread(buffer, 1, block_size, input_des);
                 bytes_read += read_count;
 
+                if (read_count == 0) break;
+
                 s32 new_size = bz3_encode_block(state, buffer, read_count);
                 if (new_size == -1) {
                     fprintf(stderr, "Failed to encode a block: %s\n", bz3_strerror(state));
@@ -236,6 +244,10 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
                 new_size = read_neutral_s32(byteswap_buf);
                 xread_noeof(&byteswap_buf, 1, 4, input_des);
                 old_size = read_neutral_s32(byteswap_buf);
+                if (old_size > block_size + 31) {
+                    fprintf(stderr, "Failed to decode a block: Inconsistent headers.\n");
+                    return 1;
+                }
                 xread_noeof(buffer, 1, new_size, input_des);
                 bytes_read += 8 + new_size;
                 if (bz3_decode_block(state, buffer, new_size, old_size) == -1) {
@@ -253,6 +265,10 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
                 new_size = read_neutral_s32(byteswap_buf);
                 xread_noeof(&byteswap_buf, 1, 4, input_des);
                 old_size = read_neutral_s32(byteswap_buf);
+                if (old_size > block_size + 31) {
+                    fprintf(stderr, "Failed to decode a block: Inconsistent headers.\n");
+                    return 1;
+                }
                 xread_noeof(buffer, 1, new_size, input_des);
                 bytes_read += 8 + new_size;
                 if (bz3_decode_block(state, buffer, new_size, old_size) == -1) {
@@ -326,6 +342,10 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
                     sizes[i] = read_neutral_s32(byteswap_buf);
                     xread_noeof(&byteswap_buf, 1, 4, input_des);
                     old_sizes[i] = read_neutral_s32(byteswap_buf);
+                    if (old_sizes[i] > block_size + 31) {
+                        fprintf(stderr, "Failed to decode a block: Inconsistent headers.\n");
+                        return 1;
+                    }
                     xread_noeof(buffers[i], 1, sizes[i], input_des);
                     bytes_read += 8 + sizes[i];
                 }
@@ -350,6 +370,10 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
                     sizes[i] = read_neutral_s32(byteswap_buf);
                     xread_noeof(&byteswap_buf, 1, 4, input_des);
                     old_sizes[i] = read_neutral_s32(byteswap_buf);
+                    if (old_sizes[i] > block_size + 31) {
+                        fprintf(stderr, "Failed to decode a block: Inconsistent headers.\n");
+                        return 1;
+                    }
                     xread_noeof(buffers[i], 1, sizes[i], input_des);
                     bytes_read += 8 + sizes[i];
                 }
@@ -370,14 +394,16 @@ static int process(FILE * input_des, FILE * output_des, int mode, int block_size
     }
 #endif
 
-    if(verbose) {
-        if(file_name) fprintf(stderr, " %s:", file_name);
-        if(mode == MODE_ENCODE)
-            fprintf(stderr, "\t%"PRIu64" -> %"PRIu64" bytes, %.2f%%, %.2f bpb\n", bytes_read, bytes_written, (double)bytes_written * 100.0 / bytes_read, (double)bytes_written * 8.0 / bytes_read);
-        else if(mode == MODE_DECODE)
-            fprintf(stderr, "\t%"PRIu64" -> %"PRIu64" bytes, %.2f%%, %.2f bpb\n", bytes_read, bytes_written, (double)bytes_read * 100.0 / bytes_written, (double)bytes_read * 8.0 / bytes_written);
+    if (verbose) {
+        if (file_name) fprintf(stderr, " %s:", file_name);
+        if (mode == MODE_ENCODE)
+            fprintf(stderr, "\t%" PRIu64 " -> %" PRIu64 " bytes, %.2f%%, %.2f bpb\n", bytes_read, bytes_written,
+                    (double)bytes_written * 100.0 / bytes_read, (double)bytes_written * 8.0 / bytes_read);
+        else if (mode == MODE_DECODE)
+            fprintf(stderr, "\t%" PRIu64 " -> %" PRIu64 " bytes, %.2f%%, %.2f bpb\n", bytes_read, bytes_written,
+                    (double)bytes_read * 100.0 / bytes_written, (double)bytes_read * 8.0 / bytes_written);
         else
-            fprintf(stderr, "OK, %"PRIu64" bytes read.\n", bytes_read);
+            fprintf(stderr, "OK, %" PRIu64 " bytes read.\n", bytes_read);
     }
 
     return 0;
@@ -489,7 +515,8 @@ int main(int argc, char * argv[]) {
             case '?':
                 fprintf(stderr, "Try 'bzip3 --help' for more information.\n");
                 return 1;
-            case 'e': case 'z':
+            case 'e':
+            case 'z':
                 mode = MODE_ENCODE;
                 break;
             case 'd':
@@ -685,6 +712,8 @@ int main(int argc, char * argv[]) {
     output_des = mode != MODE_TEST ? open_output(output, force) : NULL;
     input_des = open_input(input);
 
+    if (output != f2) free(output);
+
     int r = process(input_des, output_des, mode, block_size, workers, verbose, input);
 
     fclose(input_des);
